/*
 * Copyright (C) 2010, Sasa Zivkov <sasa.zivkov@sap.com>
 * and other copyright owners as documented in the project's IP log.
 *
 * This program and the accompanying materials are made available
 * under the terms of the Eclipse Distribution License v1.0 which
 * accompanies this distribution, is reproduced below, and is
 * available at http://www.eclipse.org/org/documents/edl-v10.php
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or
 * without modification, are permitted provided that the following
 * conditions are met:
 *
 * - Redistributions of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 *
 * - Redistributions in binary form must reproduce the above
 *   copyright notice, this list of conditions and the following
 *   disclaimer in the documentation and/or other materials provided
 *   with the distribution.
 *
 * - Neither the name of the Eclipse Foundation, Inc. nor the
 *   names of its contributors may be used to endorse or promote
 *   products derived from this software without specific prior
 *   written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package org.eclipse.jgit.notes;

import java.io.IOException;
import java.text.MessageFormat;

import org.eclipse.jgit.internal.JGitText;

/**
 * This exception will be thrown from the
 * {@link org.eclipse.jgit.notes.NoteMerger} when a conflict on Notes content is
 * found during merge.
 */
public class NotesMergeConflictException extends IOException {
	private static final long serialVersionUID = 1L;

	/**
	 * Construct a NotesMergeConflictException for the specified base, ours and
	 * theirs note versions.
	 *
	 * @param base
	 *            note version
	 * @param ours
	 *            note version
	 * @param theirs
	 *            note version
	 */
	public NotesMergeConflictException(Note base, Note ours, Note theirs) {
		super(MessageFormat.format(JGitText.get().mergeConflictOnNotes,
				noteOn(base, ours, theirs), noteData(base), noteData(ours),
				noteData(theirs)));
	}

	/**
	 * Constructs a NotesMergeConflictException for the specified base, ours and
	 * theirs versions of the root note tree.
	 *
	 * @param base
	 *            version of the root note tree
	 * @param ours
	 *            version of the root note tree
	 * @param theirs
	 *            version of the root note tree
	 */
	public NotesMergeConflictException(NonNoteEntry base, NonNoteEntry ours,
			NonNoteEntry theirs) {
		super(MessageFormat.format(
				JGitText.get().mergeConflictOnNonNoteEntries, name(base),
				name(ours), name(theirs)));
	}

	private static String noteOn(Note base, Note ours, Note theirs) {
		if (base != null)
			return base.name();
		if (ours != null)
			return ours.name();
		return theirs.name();
	}

	private static String noteData(Note n) {
		if (n != null)
			return n.getData().name();
		return ""; //$NON-NLS-1$
	}

	private static String name(NonNoteEntry e) {
		return e != null ? e.name() : ""; //$NON-NLS-1$
	}
}
